#ifndef FRAMECPP__COMMON__FrameStreamWrapper_HH
#define FRAMECPP__COMMON__FrameStreamWrapper_HH

#include "framecpp/Common/FrameStream.hh"

namespace FrameCPP
{
  namespace Common
  {
#define DECL INT_2U SPEC, typename AdcData, typename FrameH, typename ProcData, typename SerData, typename SimData, typename RawData, typename Event, typename SimEvent
#define DECL_PARAMS SPEC, AdcData, FrameH, ProcData, SerData, SimData, RawData, Event, SimEvent

    //-------------------------------------------------------------------
    /// \brief Wrapper to properly cast return values into the namespace
    ///        associated with the frame specification namespace.
    //-------------------------------------------------------------------
    template< DECL >
    class IFrameStreamWrapper
      : public IFrameStream
    {
    public:
      //-----------------------------------------------------------------
      /// \brief Definition of buffer_type.
      //-----------------------------------------------------------------
      typedef IFrameStream::buffer_type buffer_type;

      //-----------------------------------------------------------------
      /// \brief Constructor
      //-----------------------------------------------------------------
      IFrameStreamWrapper( buffer_type* Buffer );

      //-----------------------------------------------------------------
      /// \brief Extract the requested FrameH structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] ContainerSet
      ///     Bit map of components to included.
      ///
      /// \return
      ///     Upon successful completion, the address of the FrameH
      ///     structure linked with all requested subcomponents
      ///     by ContainerSet.
      //-----------------------------------------------------------------
      FrameH* ReadFrameH( INT_4U Frame, INT_4U ContainerSet );

      //-----------------------------------------------------------------
      /// \brief Extract the requested FrAdcData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Name of the channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrAdcData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      AdcData* ReadFrAdcData( INT_4U Frame,
			      const std::string& Channel );
      //-----------------------------------------------------------------
      /// \brief Extract the requested FrAdcData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Zero based index of channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrAdcData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      AdcData* ReadFrAdcData( INT_4U Frame, INT_4U Channel );
      //-----------------------------------------------------------------
      /// \brief Extract the requested FrAdcData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Name of the channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrAdcData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      AdcData* ReadFrAdcStruct( INT_4U Frame, const std::string& Channel );
      //-----------------------------------------------------------------
      /// \brief Extract the requested FrAdcData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Zero based index of channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrAdcData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      AdcData* ReadFrAdcStruct( INT_4U Frame, INT_4U Channel );


      //-----------------------------------------------------------------
      /// \brief Extract the requested FrEvent structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Name of the channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrEvent is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      Event* ReadFrEvent( INT_4U Frame, const std::string& Channel );

      //-----------------------------------------------------------------
      /// \brief Extract the requested FrEvent structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Name of the channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrEvent is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      Event* ReadFrEventStruct( INT_4U Frame, const std::string& Channel );

      //-----------------------------------------------------------------
      /// \brief Extract the requested FrProcData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Name of the channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrProcData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      ProcData* ReadFrProcData( INT_4U Frame, const std::string& Channel );
      //-----------------------------------------------------------------
      /// \brief Extract the requested FrProcData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Zero based index of channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrProcData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      ProcData* ReadFrProcData( INT_4U Frame, INT_4U Channel );
      //-----------------------------------------------------------------
      /// \brief Extract the requested FrProcData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Name of the channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrProcData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      ProcData* ReadFrProcStruct( INT_4U Frame, const std::string& Channel );

      //-----------------------------------------------------------------
      /// \brief Extract the requested FrProcData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Zero based index of channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrProcData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      ProcData* ReadFrProcStruct( INT_4U Frame, INT_4U Channel );

      //-----------------------------------------------------------------
      /// \brief Extract the requested FrRawData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrRawData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      RawData* ReadFrRawData( INT_4U Frame );

      //-----------------------------------------------------------------
      /// \brief Extract the requested FrSerData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Name of the channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrSerData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      SerData* ReadFrSerData( INT_4U Frame, const std::string& Channel );

      //-----------------------------------------------------------------
      /// \brief Extract the requested FrSerData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Name of the channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrSerData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      SerData* ReadFrSerStruct( INT_4U Frame, const std::string& Channel );

      //-----------------------------------------------------------------
      /// \brief Extract the requested FrSimEvent structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Name of the channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrSimEvent is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      SimEvent* ReadFrSimEvent( INT_4U Frame, const std::string& Channel );

      //-----------------------------------------------------------------
      /// \brief Extract the requested FrSimEvent structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Name of the channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrSimEvent is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      SimEvent* ReadFrSimEventStruct( INT_4U Frame, const std::string& Channel );

      //-----------------------------------------------------------------
      /// \brief Extract the requested FrSerData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Name of the channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrSimData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      SimData* ReadFrSimStruct( INT_4U Frame, const std::string& Channel );
      //-----------------------------------------------------------------
      /// \brief Extract the requested FrSerData structure from the stream
      ///
      /// \param[in] Frame
      ///     Zero based index of the frame.
      /// \param[in] Channel
      ///     Name of the channel being requested.
      ///
      /// \return
      ///     Upon success, a non-NULL pointer to the FrSimData is
      ///     returned. Upon failure, either an exception is thrown
      //      or a NULL pointer is returned.
      //-----------------------------------------------------------------
      SimData* ReadFrSimData( INT_4U Frame, const std::string& Channel );

      //-----------------------------------------------------------------
      /// \brief Extract the next FrameH structure from the stream
      ///
      /// \return
      ///     Upon success, a pointer to a FrameH structure.
      //-----------------------------------------------------------------
      FrameH* ReadNextFrame( );
    };

    template< DECL >
    IFrameStreamWrapper< DECL_PARAMS >::
    IFrameStreamWrapper( buffer_type* Buffer )
      : IFrameStream( Buffer, SPEC )
    {
    }

    template< DECL >
    FrameH* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrameH( INT_4U Frame, INT_4U ContainerSet )
    {
      return
	dynamic_cast< FrameH* >( readFrameHSubset( Frame, ContainerSet ) );
    }

    template< DECL >
    AdcData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrAdcData( INT_4U Frame, const std::string& Channel )
    {
      return
	dynamic_cast< AdcData* >( readFrAdcData( Frame, Channel ) );
    }
    
    template< DECL >
    AdcData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrAdcData( INT_4U Frame, INT_4U Channel )
    {
      return
	dynamic_cast< AdcData* >( readFrAdcData( Frame, Channel ) );
    }
    
    template< DECL >
    AdcData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrAdcStruct( INT_4U Frame, const std::string& Channel )
    {
      return
	dynamic_cast< AdcData* >( readFrAdcStruct( Frame, Channel ) );
    }
    
    template< DECL >
    AdcData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrAdcStruct( INT_4U Frame, INT_4U Channel )
    {
      return
	dynamic_cast< AdcData* >( readFrAdcStruct( Frame, Channel ) );
    }
    
    template< DECL >
    Event* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrEvent( INT_4U Frame, const std::string& Channel )
    {
      return
	dynamic_cast< Event* >( readFrEvent( Frame, Channel ) );
    }
    
    template< DECL >
    Event* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrEventStruct( INT_4U Frame, const std::string& Channel )
    {
      return
	dynamic_cast< Event* >( readFrEventStruct( Frame, Channel ) );
    }
    
    template< DECL >
    ProcData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrProcData( INT_4U Frame, const std::string& Channel )
    {
      return
	dynamic_cast< ProcData* >( readFrProcData( Frame, Channel ) );
    }
    
    template< DECL >
    ProcData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrProcData( INT_4U Frame, INT_4U Channel )
    {
      return
	dynamic_cast< ProcData* >( readFrProcData( Frame, Channel ) );
    }
    
    template< DECL >
    RawData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrRawData( INT_4U Frame )
    {
      return
	dynamic_cast< RawData* >( readFrRawData( Frame ) );
    }

    template< DECL >
    ProcData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrProcStruct( INT_4U Frame, const std::string& Channel )
    {
      return
	dynamic_cast< ProcData* >( readFrProcStruct( Frame, Channel ) );
    }
    
    template< DECL >
    ProcData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrProcStruct( INT_4U Frame, INT_4U Channel )
    {
      return
	dynamic_cast< ProcData* >( readFrProcStruct( Frame, Channel ) );
    }
    
    template< DECL >
    SerData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrSerData( INT_4U Frame, const std::string& Channel )
    {
      return
	dynamic_cast< SerData* >( readFrSerData( Frame, Channel ) );
    }
    
    template< DECL >
    SerData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrSerStruct( INT_4U Frame, const std::string& Channel )
    {
      return
	dynamic_cast< SerData* >( readFrSerStruct( Frame, Channel ) );
    }
    
    template< DECL >
    SimData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrSimData( INT_4U Frame, const std::string& Channel )
    {
      return
	dynamic_cast< SimData* >( readFrSimData( Frame, Channel ) );
    }
    
    template< DECL >
    SimData* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrSimStruct( INT_4U Frame, const std::string& Channel )
    {
      return
	dynamic_cast< SimData* >( readFrSimStruct( Frame, Channel ) );
    }
    
    template< DECL >
    SimEvent* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrSimEvent( INT_4U Frame, const std::string& Channel )
    {
      return
	dynamic_cast< SimEvent* >( readFrSimEvent( Frame, Channel ) );
    }
    
    template< DECL >
    SimEvent* IFrameStreamWrapper< DECL_PARAMS >::
    ReadFrSimEventStruct( INT_4U Frame, const std::string& Channel )
    {
      return
	dynamic_cast< SimEvent* >( readFrSimEventStruct( Frame, Channel ) );
    }
    
    template< DECL >
    FrameH* IFrameStreamWrapper< DECL_PARAMS >::
    ReadNextFrame( )
    {
      return
	dynamic_cast< FrameH* >( readNextFrame( ) );
    }
    
#undef DECL_PARAMS
#undef DECL

  } // namespace - Common
} // namespace - FrameCPP

#endif /* FRAMECPP__COMMON__FrameStreamWrapper_HH */
