#ifndef FrameCPP_VERSION_6_FrDetector_HH
#define FrameCPP_VERSION_6_FrDetector_HH

#include <string>
#include <vector>

#include "general/types.hh"

#include "framecpp/FrameCPP.hh"

#include "framecpp/Common/SearchContainer.hh"
#include "framecpp/Common/TOCInfo.hh"
#include "framecpp/Common/FrDetector.hh"

#include "framecpp/Version6/GPSTime.hh"
#include "framecpp/Version6/STRING.hh"

#include "framecpp/Version6/FrTable.hh"
#include "framecpp/Version6/FrVect.hh"

#include "framecpp/Version4/FrDetector.hh"

namespace FrameCPP
{
  namespace Version_6
  {
    class FrDetector
      : public Common::FrDetector,
	public Common::TOCInfo
    {
    public:
      typedef Common::SearchContainer< FrVect, &FrVect::GetName >
      aux_type;
      typedef aux_type::iterator aux_iterator;
      typedef aux_type::const_iterator const_aux_iterator;

      typedef Common::SearchContainer< FrTable, &FrTable::GetNameSlow >
      table_type;
      typedef table_type::iterator table_iterator;
      typedef table_type::const_iterator const_table_iterator;

      //: Default constructor
      FrDetector( );

      FrDetector( const std::string& Name,
		  const char* Prefix,
		  const REAL_8 Longitude,
		  const REAL_8 Latitude,
		  const REAL_4 Elevation,
		  const REAL_4 ArmXazimuth,
		  const REAL_4 ArmYazimuth,
		  const REAL_4 ArmXaltitude,
		  const REAL_4 ArmYaltitude,
		  const REAL_4 ArmXmidpoint,
		  const REAL_4 ArmYmidpoint,
		  const INT_4S LocalTime );

      FrDetector( const FrDetector& detector );

      FrDetector( Previous::FrDetector& Source,
		  Common::IStream* Stream );

      static const char* const StructName( );

      static const Common::Description* StructDescription( );

      //: Number of bytes needed to write this structure
      cmn_streamsize_type Bytes( const Common::StreamBase& Stream) const;

      virtual FrDetector* Create( Common::IStream& Stream ) const;

      virtual const char* const ObjectStructName( ) const;

      virtual void Write( Common::OStream& Stream ) const;

      const std::string& GetName( ) const;
      const CHAR* GetPrefix( ) const { return m_data.prefix; }
      REAL_8 GetLongitude( ) const { return m_data.longitude; }
      REAL_8 GetLatitude( ) const { return m_data.latitude; }
      REAL_4 GetElevation( ) const { return m_data.elevation; }
      REAL_4 GetArmXazimuth( ) const { return m_data.armXazimuth; }
      REAL_4 GetArmYazimuth( ) const { return m_data.armYazimuth; }
      REAL_4 GetArmXaltitude( ) const { return m_data.armXaltitude; }
      REAL_4 GetArmYaltitude( ) const { return m_data.armYaltitude; }
      REAL_4 GetArmXmidpoint( ) const { return m_data.armXmidpoint; }
      REAL_4 GetArmYmidpoint( ) const { return m_data.armYmidpoint; }
      INT_4S GetLocalTime( ) const { return m_data.localTime; }
      const aux_type& RefAux( ) const { return m_data.aux; }
      aux_type& RefAux( ) { return m_data.aux; }
      const table_type& RefTable( ) const { return m_data.table; }
      table_type& RefTable( ) { return m_data.table; }   
   
      //: Merge with another FrDetector
      FrDetector& Merge( const FrDetector& RHS );

      static std::auto_ptr< Object >
      Promote( INT_2U Source,
	       Object* Obj, Common::IStream* Stream );

      /// \brief comparison operator
      bool operator==( const FrDetector& RHS ) const;

      /// \brief comparison operator
      virtual bool operator==( const Common::FrameSpec::Object& Obj ) const;

      virtual void TOCQuery( int InfoClass, ... ) const;

    protected:
      /// \brief Down grade an object
      virtual std::auto_ptr< Object >
      demote( INT_2U Target, Common::IStream* Stream ) const;

      /// \brief Upgrade an object
      virtual std::auto_ptr< Object >
      promote( INT_2U Target, Object* Obj, Common::IStream* Stream ) const;

    private:
      using Common::FrameSpec::Object::Create;

      struct fr_detector_data_type {
	bool operator==( const fr_detector_data_type& RHS ) const;
	STRING		name;
	CHAR		prefix[2];
	REAL_8		longitude;
	REAL_8		latitude;
	REAL_4		elevation;
	REAL_4		armXazimuth;
	REAL_4		armYazimuth;
	REAL_4		armXaltitude;
	REAL_4		armYaltitude;
	REAL_4		armXmidpoint;
	REAL_4		armYmidpoint;
	INT_4S		localTime;
	aux_type	aux;
	table_type	table;	
      };

      fr_detector_data_type	m_data;

      FrDetector( IStream& Stream );
    };

    inline std::auto_ptr< Common::FrameSpec::Object > FrDetector::
    Promote( INT_2U Source,
	     Common::FrameSpec::Object* Obj, Common::IStream* Stream )
    {
      return Object::PromoteObject< Previous::FrDetector, FrDetector >
	( DATA_FORMAT_VERSION, Source, Obj, Stream );
    }

    inline const char* const FrDetector::
    StructName( )
    {
      static const char* const name = "FrDetector";
      return name;
    }

    inline bool FrDetector::
    operator==( const FrDetector& RHS ) const
    {
      return ( m_data == RHS.m_data );
    }
  } // namespace - Version_6
} // namespace FrameCPP
#endif /* FrameCPP_VERSION_6_FrDetector_HH */
